/*
 * Berkeley Lab Checkpoint/Restart (BLCR) for Linux is Copyright (c)
 * 2003, The Regents of the University of California, through Lawrence
 * Berkeley National Laboratory (subject to receipt of any required
 * approvals from the U.S. Dept. of Energy).  All rights reserved.
 *
 * Portions may be copyrighted by others, as may be noted in specific
 * copyright notices within specific files.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <errno.h>
#include <sys/types.h>
#include <unistd.h>
#include <mpi.h>
#include <string.h>
#include <signal.h>

static int limit = 120; /* default */
static int count = 0;
static int np = 0;
char processor_name[MPI_MAX_PROCESSOR_NAME];

static void delay(void)
{
    struct timespec ts = {tv_sec: 1, tv_nsec: 0};
    int rc;

    do {
        rc = nanosleep(&ts, &ts);
    } while ((rc <0) && (errno == EINTR) && (ts.tv_sec || ts.tv_nsec));
}

static void advance(void)
{
    ++count;
}

static void broadcast(int *buf)
{
    MPI_Bcast(buf, 1, MPI_INT, 0, MPI_COMM_WORLD);
}

static void output(int id)
{
    if ((count % np) == id) {
        printf("Processor %s says count = %d\n", processor_name, count);
    }

}

static void thread(int id)
{
    do {
        broadcast(&count);
        output(id);
        delay();
        advance();
    } while (count < limit);
}

int main(int argc, char *argv[], char *envp[])
{
    int rank;
    int namelen;
    int i;
    const char count_killpid[] = "COUNT_KILLPID";
    char *killpidstr = NULL;
    int killpid = 0;

    MPI_Init(&argc, &argv);
    MPI_Comm_rank(MPI_COMM_WORLD, &rank);
    MPI_Comm_size(MPI_COMM_WORLD, &np);
    MPI_Get_processor_name(processor_name, &namelen);

    killpidstr = getenv(count_killpid);
    if (killpidstr != NULL) {
        killpid = atoi(killpidstr);
    }

    if (rank == 0) {
        printf("Starting main process w/ pid %d", (int) getpid());

        if (killpid) {
            kill(killpid, SIGUSR1); 
            printf(", notifying %d.\n", killpid);
        } else {
            printf(".\n");
        }
    }

    if (argc > 1) {
        limit = (i = atoi(argv[1])) > 0 ? i : limit;
    }

    broadcast(&limit);

    thread(rank);

    MPI_Finalize();

    return 0;
}
